import click
from collections import OrderedDict

from .colors import CYAN


UP_ONE = "\033[A\033[1000D"
CLEAR_LINE = "\033[2K"


class ProgressPrinter:
    """
    Prints multiple in-progresses statuses to the terminal in a way where they
    rewrite their own lines.
    """

    prefix = ""

    def __init__(self):
        self.current = OrderedDict()

    def start(self, noun, verb, status=""):
        """
        Adds something to be displayed as "verb noun: status".
        Noun is unique; if you re-use it, you must end() it before you start()
        the same noun again.
        """
        assert noun not in self.current
        # Get into the right place on the terminal
        print(UP_ONE * len(self.current), flush=True, end="")
        # Add our noun in
        self.current[noun] = {"verb": verb, "status": status}
        # Re-print the statuses
        self._print_all()

    def update(self, noun, status):
        """
        Updates what's printed after "verb noun".
        """
        self.current[noun]['status'] = status
        self._print_all()

    def end(self, noun, status=None):
        """
        Removes the item from the live list and pushes it above with a final status.
        """
        if status is not None:
            self.current[noun]['status'] = status
        # Get into the right place on the terminal
        print(UP_ONE * len(self.current), flush=True, end="")
        # Print the single item we're removing
        self._print_one(noun)
        # Remove it from the rest and print
        del self.current[noun]
        self._print_all()

    def _print_all(self):
        """
        Prints all items.
        """
        for noun in self.current.keys():
            self._print_one(noun)

    def _print_one(self, noun):
        details = self.current[noun]
        print(CLEAR_LINE + self.prefix + "{} {}: {}".format(details['verb'], CYAN(noun), details['status']))


class RunProgressPrinter(ProgressPrinter):
    """
    Translates run/stop progress notifications into nice console output.
    """

    def __call__(self, progtype, data):
        handler = getattr(self, "prog_{}".format(progtype), None)
        if handler:
            handler(data)
        else:
            self.prog_default(progtype, data)

    def prog_to_stop(self, data):
        if data['instances']:
            click.echo(
                self.prefix +
                "Plan to stop: " +
                ", ".join(sorted(CYAN(instance.container.name) for instance in data['instances']))
            )

    def prog_to_start(self, data):
        if data['instances']:
            click.echo(
                self.prefix +
                "Plan to start: " +
                ", ".join(sorted(CYAN(instance.container.name) for instance in data['instances']))
            )

    def prog_stop_begin(self, data):
        self.start(data['instance'].container.name, "Stopping")

    def prog_stop_end(self, data):
        self.end(data['instance'].container.name, "Done")

    def prog_start_begin(self, data):
        self.start(data['instance'].container.name, "Starting")

    def prog_start_progress(self, data):
        self.update(data['instance'].container.name, data['message'])

    def prog_start_end(self, data):
        self.end(data['instance'].container.name, "Done")

    def prog_default(self, progtype, data):
        click.echo("{}: {}".format(progtype, data))



import difflib

try:
    import pylev
except ImportError:
    pylev = None


if pylev:
    def spell_correct(input, choices, threshold=1.0):
        """
        Find a possible spelling correction for a given input.
        """
        # Try to find corrections using pylev
        guesses = sorted((pylev.levenshtein(c, input), c) for c in choices)
        if not guesses:
            return None
        distance, suggestion = next(iter(guesses))
        # Score the error distance based on word length
        length = max(len(suggestion), len(input))
        score = distance ** 2 / length
        if score <= threshold:
            return suggestion
else:
    def spell_correct(input, choices, threshold=0.6):
        """
        Find a possible spelling correction for a given input.
        """
        guesses = difflib.get_close_matches(input, choices, 1, cutoff=threshold)
        return next(iter(guesses), None)



import attr
import click
import sys

from .base import BasePlugin
from ..cli.argument_types import HostType
from ..cli.tasks import Task


@attr.s
class DoctorPlugin(BasePlugin):
    """
    System for running sets of checks defined in plugins to establish system health.
    """

    provides = ["doctor"]

    def load(self):
        self.add_command(doctor)
        self.add_catalog_type("doctor-exam")


@attr.s
class BaseExamination:
    """
    Base class for doctor examinations.

    Write one or more methods that start with check_ and each one will be run
    as a separate exam, with their docstring as the name of the exam.

    If a method raises Failure or Warning, they're printed out to the console.
    If it returns without raising, it is considered successful.
    """

    # Class-level attribute for the check description
    description = None

    app = attr.ib()

    class Failure(BaseException):
        """
        Raised by checks on failure.
        """

    class Warning(BaseException):
        """
        Raised by checks on non-failure, but worrying info.
        """

    def skipped(self):
        return False

    def run(self, host, parent_task):
        """
        Runs through examinations if we're not being skipped.
        """
        # Check skipping
        if self.skipped():
            return
        # Make overall task
        self.host = host
        assert self.description is not None, "You must override description on an examination subclass"
        task = Task(self.description, parent=parent_task)
        # Find all methods
        methods = []
        for name in sorted(dir(self)):
            if name.startswith("check_"):
                methods.append(getattr(self, name))
        if not methods:
            task.finish(status="Empty", status_flavor=Task.FLAVOR_WARNING)
            return
        # Run them in alphabetical order
        for method in methods:
            method_description = method.__doc__.strip() or method.__name__
            subtask = Task(method_description, parent=task)
            try:
                method()
            except self.Warning as e:
                subtask.set_extra_info([x.strip() for x in str(e).split("\n")])
                subtask.finish(status="Warning", status_flavor=Task.FLAVOR_WARNING)
            except self.Failure as e:
                subtask.set_extra_info([x.strip() for x in str(e).split("\n")])
                subtask.finish(status="Failed", status_flavor=Task.FLAVOR_BAD)
                return False
            else:
                subtask.finish(status="OK", status_flavor=Task.FLAVOR_GOOD)
        task.finish(status="OK", status_flavor=Task.FLAVOR_GOOD)
        return True


@click.command()
@click.option('--host', '-h', type=HostType(), default='default')
@click.pass_obj
def doctor(app, host):
    """
    Runs the doctor exams.
    """
    exams = app.get_catalog_items("doctor-exam").values()
    problem = False
    for exam in exams:
        if not exam(app).run(host, app.root_task):
            problem = True
    if problem:
        sys.exit(1)



import datetime
import io
import json
import logging
import os
import tarfile
import tempfile

import attr
from docker.utils import exclude_paths

from ..cli.colors import CYAN, remove_ansi
from ..cli.tasks import Task

from ..exceptions import BuildFailureError, FailedCommandException


class TaskExtraInfoHandler(logging.Handler):
    """
    Custom log handler that emits to a task's extra info.
    """

    def __init__(self, task):
        super(TaskExtraInfoHandler, self).__init__()
        self.task = task

    def emit(self, record):
        text = self.format(record)
        # Sanitise the text and make it short-ish
        text = remove_ansi(text).replace("\n", "").replace("\r", "").strip()[:80]
        self.task.set_extra_info(
            self.task.extra_info[-3:] + [text]
        )


@attr.s
class Builder:
    """
    Build an image from a single container.
    """
    host = attr.ib()
    container = attr.ib()
    app = attr.ib()
    logfile_name = attr.ib()
    parent_task = attr.ib()
    # Set docker_cache to False to force docker to rebuild every layer.
    docker_cache = attr.ib(default=True)
    verbose = attr.ib(default=False)
    logger = attr.ib(init=False)

    def __attrs_post_init__(self):
        self.logger = logging.getLogger('build_logger')
        self.logger.setLevel(logging.INFO)

        # Close all old logging handlers
        if self.logger.handlers:
            [handler.close() for handler in self.logger.handlers]
            self.logger.handlers = []

        # Add build log file handler
        file_handler = logging.FileHandler(self.logfile_name)
        self.logger.addHandler(file_handler)

        # Optionally add task (console) log handler
        self.task = Task(
            "Building {}".format(CYAN(self.container.name)),
            parent=self.parent_task,
            collapse_if_finished=True,
        )
        if self.verbose:
            self.logger.addHandler(TaskExtraInfoHandler(self.task))

    def build(self):
        """
        Runs the build process and raises BuildFailureError if it fails.
        """
        self.logger.info("Building image {}".format(self.container.name))

        build_successful = True
        progress = 0
        start_time = datetime.datetime.now().replace(microsecond=0)

        self.app.run_hooks('pre-build', host=self.host, container=self.container, task=self.task)

        try:
            # Prep normalised context
            build_context = self.make_build_context()
            # Run build
            result = self.host.client.build(
                self.container.path,
                dockerfile=self.container.dockerfile_name,
                tag=self.container.image_name,
                nocache=not self.docker_cache,
                rm=True,
                stream=True,
                custom_context=True,
                encoding="gzip",
                fileobj=build_context,
                buildargs=self.container.buildargs,
                # If the parent image is not in prefix, pull it during build
                pull=not self.container.build_parent_in_prefix,
            )
            with self.task.rate_limit() as limited_task:
                self.logger.task = limited_task
                for data in result:
                    # Make sure data is a string
                    if isinstance(data, bytes):
                        data = data.decode("utf8")
                    # Deal with any potential double chunks
                    for data_segment in data.strip().split("\r\n"):
                        data_obj = json.loads(data_segment.strip())
                        if 'stream' in data_obj:
                            # docker data stream has extra newlines in it, so we will
                            # strip them before logging.
                            self.logger.info(data_obj['stream'].rstrip())
                            if data_obj['stream'].startswith('Step '):
                                progress += 1
                                self.task.update(status="." * progress)
                        if 'error' in data_obj:
                            self.logger.info(data_obj['error'].rstrip())
                            build_successful = False
                self.logger.task = self.task

            if not build_successful:
                raise FailedCommandException

        except FailedCommandException:
            message = "Build FAILED for image {}!".format(self.container.name)
            self.logger.info(message)
            self.task.finish(status="FAILED", status_flavor=Task.FLAVOR_BAD)
            raise BuildFailureError(message)

        else:
            # Run post-build hooks
            self.app.run_hooks('post-build', host=self.host, container=self.container, task=self.task)

            # Print out end-of-build message
            end_time = datetime.datetime.now().replace(microsecond=0)
            time_delta_str = str(end_time - start_time)
            if time_delta_str.startswith('0:'):
                # no point in showing hours, unless it runs for more than one hour
                time_delta_str = time_delta_str[2:]

            build_completion_message = "Build time for {image_name} image: {build_time}".format(
                image_name=self.container.name,
                build_time=time_delta_str
            )
            self.logger.info(build_completion_message)

            # Close out the task
            self.task.finish(status='Done [{}]'.format(time_delta_str), status_flavor=Task.FLAVOR_GOOD)

    def make_build_context(self):
        """
        Makes a Docker build context from a local directory.
        Normalises all file ownership and times so that the docker hashes align
        better.
        """
        # Start temporary tar file
        fileobj = tempfile.NamedTemporaryFile()
        tfile = tarfile.open(mode='w:gz', fileobj=fileobj)
        # Get list of files/dirs to add to the tar
        paths = exclude_paths(self.container.path, [])
        # For each file, add it to the tar with normalisation
        for path in paths:
            disk_location = os.path.join(self.container.path, path)
            # Directory addition
            if os.path.isdir(disk_location):
                info = tarfile.TarInfo(name=path)
                info.mtime = 0
                info.mode = 0o775
                info.type = tarfile.DIRTYPE
                info.uid = 0
                info.gid = 0
                info.uname = "root"
                info.gname = "root"
                tfile.addfile(info)
            # Normal file addition
            elif os.path.isfile(disk_location):
                stat = os.stat(disk_location)
                info = tarfile.TarInfo(name=path)
                info.mtime = 0
                info.size = stat.st_size
                info.mode = 0o755
                info.type = tarfile.REGTYPE
                info.uid = 0
                info.gid = 0
                info.uname = "root"
                info.gname = "root"
                # Rewrite docker FROM lines with a : in them and raise a warning
                # TODO: Deprecate this!
                if path.lstrip("/") == self.container.dockerfile_name:
                    # Read in dockerfile line by line, replacing the FROM line
                    dockerfile = io.BytesIO()
                    with open(disk_location, "r") as fh:
                        for line in fh:
                            if line.upper().startswith("FROM ") and self.container.build_parent_in_prefix:
                                line = line.replace(":", "-")
                            dockerfile.write(line.encode("utf8"))
                    dockerfile.seek(0)
                    tfile.addfile(info, dockerfile)
                else:
                    with open(disk_location, "rb") as fh:
                        tfile.addfile(info, fh)
            # Error for anything else
            else:
                raise ValueError(
                    "Cannot add non-file/dir %s to docker build context" % path
                )
        # Return that tarfile
        tfile.close()
        fileobj.seek(0)
        return fileobj



import attr
import click
import datetime
import sys

from .base import BasePlugin
from ..cli.colors import CYAN, GREEN, RED, remove_ansi
from ..cli.argument_types import ContainerType, HostType
from ..cli.tasks import Task
from ..docker.build import Builder
from ..exceptions import BuildFailureError, ImagePullFailure
from ..utils.sorting import dependency_sort
from ..constants import PluginHook


@attr.s
class BuildPlugin(BasePlugin):
    """
    Plugin for showing information about containers
    """

    provides = ["build"]

    def load(self):
        self.add_command(build)
        self.add_catalog_type("registry")


@click.command()
@click.argument('containers', type=ContainerType(profile=True), nargs=-1)
@click.option('--host', '-h', type=HostType(), default='default')
@click.option('--cache/--no-cache', default=True)
@click.option('--recursive/--one', '-r/-1', default=True)
@click.option('--verbose/--quiet', '-v/-q', default=True)
# TODO: Add a proper requires_docker check
# TODO: Add build profile
@click.pass_obj
def build(app, containers, host, cache, recursive, verbose):
    """
    Build container images, along with its build dependencies.
    """
    logfile_name = app.config.get_path('bay', 'build_log_path', app)
    containers_to_pull = []
    containers_to_build = []
    pulled_containers = set()
    failed_pulls = set()

    task = Task("Building", parent=app.root_task)
    start_time = datetime.datetime.now().replace(microsecond=0)

    # Go through the containers, expanding "ContainerType.Profile" into a list
    # of default boot containers in the profile.
    for container in containers:
        if container is ContainerType.Profile:
            for con in app.containers:
                if app.containers.options(con).get('in_profile'):
                    containers_to_pull.append(con)
        else:
            containers_to_build.append(container)

    # Expand containers_to_pull (At this point just the default boot containers
    # from profile) to include runtime dependencies.
    containers_to_pull = dependency_sort(containers_to_pull, app.containers.dependencies)

    # Try pulling each container to pull, and add it to containers_to_build if
    # it fails. If it works, remember we pulled it, so we don't have to pull it
    # again later.
    for container in containers_to_pull:
        try:
            host.images.pull_image_version(
                app,
                container.image_name,
                container.image_tag,
                parent_task=task,
                fail_silently=False,
            )
        except ImagePullFailure:
            failed_pulls.add(container)
            containers_to_build.append(container)
        else:
            pulled_containers.add(container)

    ancestors_to_build = []
    # For each container to build, find its ancestry, trying to pull each
    # ancestor and stopping short if it works.
    for container in containers_to_build:
        # Always add `container` to final build list, even if recursive is
        # False.
        ancestors_to_build.append(container)
        if recursive:
            # We need to look at the ancestry starting from the oldest, up to
            # and not including the `container`
            ancestry = dependency_sort(containers_to_build,
                                       lambda x: [app.containers.build_parent(x)])[:-1]
            for ancestor in reversed(ancestry):
                try:
                    # If we've already attempted to pull it and failed, short
                    # circuit to failure block.
                    if ancestor in failed_pulls:
                        raise ImagePullFailure("We've already attempted to pull this image, and failed.")
                    # Check if we've pulled it already
                    if ancestor not in pulled_containers:
                        host.images.pull_image_version(
                            app,
                            ancestor.image_name,
                            ancestor.image_tag,
                            parent_task=task,
                            fail_silently=False,
                        )
                except ImagePullFailure:
                    failed_pulls.add(ancestor)
                    ancestors_to_build.insert(0, ancestor)
                else:
                    # We've pulled the current ancestor successfully, so skip
                    # all the older ancestors.
                    pulled_containers.add(ancestor)
                    break

    # Sort ancestors so we build the most depended on first.
    sorted_ancestors_to_build = dependency_sort(ancestors_to_build, lambda x: [app.containers.build_parent(x)])

    # dependency_sort would insert back the pulled containers into the ancestry
    # chain, so we only include ones that were in the list before
    ancestors_to_build = [
        container
        for container in sorted_ancestors_to_build
        if container in ancestors_to_build
    ]

    task.add_extra_info(
        "Order: {order}".format(
            order=CYAN(", ".join([container.name for container in ancestors_to_build])),
        ),
    )

    app.run_hooks(PluginHook.PRE_GROUP_BUILD, host=host, containers=ancestors_to_build, task=task)

    for container in ancestors_to_build:
        image_builder = Builder(
            host,
            container,
            app,
            parent_task=task,
            logfile_name=logfile_name,
            docker_cache=cache,
            verbose=verbose,
        )
        try:
            image_builder.build()
        except BuildFailureError:
            click.echo(RED("Build failed! Last 15 lines of log:"))
            # TODO: More efficient tailing
            lines = []
            with open(logfile_name, "r") as fh:
                for line in fh:
                    lines = lines[-14:] + [line]
            for line in lines:
                click.echo("  " + remove_ansi(line).rstrip())
            click.echo("See full build log at {log}".format(log=click.format_filename(logfile_name)), err=True)
            sys.exit(1)

    task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)

    # Show total build time metric after everything is complete
    end_time = datetime.datetime.now().replace(microsecond=0)
    time_delta_str = str(end_time - start_time)
    if time_delta_str.startswith('0:'):
        # no point in showing hours, unless it runs for more than one hour
        time_delta_str = time_delta_str[2:]
    click.echo("Total build time [{}]".format(GREEN(time_delta_str)))



import socket


def tcp_is_open(host, port, timeout=None):
    """
    Returns True if a TCP port is open and listening, False otherwise.
    """
    try:
        conn_kwargs = {}
        if timeout:
            conn_kwargs['timeout'] = timeout
        conn = socket.create_connection((host, port), **conn_kwargs)
        conn.close()
        return True
    except socket.error:
        return False



from cStringIO import StringIO
import json
import os.path
import struct
import tarfile

try:
    from docker.constants import STREAM_HEADER_SIZE_BYTES
except ImportError:
    # Docker-py<1.2
    from docker.client import STREAM_HEADER_SIZE_BYTES


class DockerTarStream(object):
    def __init__(self, stream):
        # We need to read into a StringIO, because tarfile does random I/O (stupid, I know)
        file = StringIO(stream.read())
        self.tar = tarfile.open(fileobj=file)

    def __getitem__(self, name):
        return self.tar.extractfile(os.path.basename(name)).read()


class DockerJSONStream(object):
    def __init__(self, stream):
        self.stream = stream

    def __iter__(self):
        for entry in self.stream:
            for chunk in entry.split("\n"):
                if chunk.strip():
                    try:
                        data = json.loads(chunk.strip())
                    except ValueError as e:
                        raise ValueError("Could not decode docker JSON stream entry: %r (%s)" % (entry, e))
                    yield data


class DockerRawStream(object):
    def __init__(self, stream):
        sock = stream._fp.fp._sock
        sock.settimeout(None)
        self.stream = stream

    def __iter__(self):
        while True:
            header = self.stream.read(STREAM_HEADER_SIZE_BYTES)
            if not header:
                break
            _, length = struct.unpack_from('>BxxxL', header)
            data = self.stream.read(length)
            if not data:
                break
            yield data



import attr
from docker.errors import NotFound

from .base import BasePlugin
from .gc import GarbageCollector
from ..cli.tasks import Task
from ..constants import PluginHook
from ..docker.build import Builder
from ..docker.introspect import FormationIntrospector
from ..docker.runner import FormationRunner


@attr.s
class BuildVolumesPlugin(BasePlugin):
    """
    Plugin for showing information about containers
    """

    requires = ["build"]

    def load(self):
        self.add_hook(PluginHook.PRE_START, self.pre_start)
        self.add_hook(PluginHook.PRE_GROUP_BUILD, self.pre_group_build)
        self.add_hook(PluginHook.POST_BUILD, self.post_build)

    def _get_providers(self):
        providers = {}
        for container in self.app.containers:
            provides_volume = container.extra_data.get("provides-volume", None)
            if provides_volume:
                providers[provides_volume] = container
        return providers

    def pre_start(self, host, instance, task):
        """
        Safety net to stop you booting volume-providing containers normally,
        and to catch and build volume containers if they're needed
        """
        # Safety net
        if instance.container.extra_data.get("provides-volume", None):
            raise ValueError("You cannot run a volume-providing container {}".format(instance.container.name))
        # If the container has named volumes, see if they're provided by anything else
        # and if so, if they're built.
        # First, collect what volumes are provided by what containers
        providers = self._get_providers()
        # Now see if any of the volumes we're trying to add need it
        for _, name in instance.container.named_volumes.items():
            if name in providers:
                # Alright, this is one that could be provided. Does it already exist?
                try:
                    host.client.inspect_volume(name)
                except NotFound:
                    # Aha! Build it!
                    Builder(
                        host,
                        providers[name],
                        self.app,
                        parent_task=task,
                        logfile_name=self.app.config.get_path(
                            'bay',
                            'build_log_path',
                            self.app,
                        ),
                        verbose=True,
                    ).build()

    def pre_group_build(self, host, containers, task):
        """
        Build volume-providing containers for all required volumes.
        """
        providers = self._get_providers()
        volumes_to_build = set()
        for container in containers:
            for volume in container.named_volumes.values():
                if volume in providers:
                    volumes_to_build.add(volume)
        for name in volumes_to_build:
            Builder(
                host,
                providers[name],
                self.app,
                parent_task=task,
                logfile_name=self.app.config.get_path(
                    'bay',
                    'build_log_path',
                    self.app,
                ),
                verbose=True,
            ).build()

    def post_build(self, host, container, task):
        """
        Intercepts builds of volume-providing containers and unpacks them.

        Volumes are stored with the ID of the corresponding volume-providing image. This will only run the container
        to recreate the volume if the image"s ID (hash) has changed.
        """
        image_details = host.client.inspect_image(container.image_name)
        provides_volume = container.extra_data.get("provides-volume", None)

        def should_extract_volume():
            if not provides_volume:
                return False
            try:
                volume_details = host.client.inspect_volume(provides_volume)
            except NotFound:
                return True
            labels = volume_details.get("Labels") or {}
            return labels.get("build_id") != image_details["Id"]

        if should_extract_volume():
            # Stop all containers that have the volume mounted
            formation = FormationIntrospector(host, self.app.containers).introspect()
            # Keep track of instances to remove after they are stopped
            instances_to_remove = formation.get_instances_using_volume(provides_volume)
            if instances_to_remove:
                formation.remove_instances(instances_to_remove)
                stop_task = Task("Stopping containers", parent=task)
                FormationRunner(self.app, host, formation, stop_task).run()
                stop_task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)
                remove_task = Task("Removing containers", parent=task)
                for instance in instances_to_remove:
                    host.client.remove_container(instance.name)
                    remove_task.update(status="Removed {}".format(instance.name))
                remove_task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)

            # Prune any orphan stopped containers, so we don't get conflict errors
            GarbageCollector(host).gc_containers(task)

            volume_task = Task("(Re)creating volume {}".format(provides_volume), parent=task)
            # Recreate the volume with the new image ID
            try:
                host.client.remove_volume(provides_volume)
                volume_task.update(status="Removed {}. Recreating".format(provides_volume))
            except NotFound:
                volume_task.update(status="Volume {} not found. Creating")
            host.client.create_volume(provides_volume, labels={"build_id": image_details["Id"]})

            # Configure the container
            volume_mountpoints = ["/volume/"]
            volume_binds = {provides_volume: {"bind": "/volume/", "mode": "rw"}}
            container_pointer = host.client.create_container(
                container.image_name,
                detach=False,
                volumes=volume_mountpoints,
                host_config=host.client.create_host_config(
                    binds=volume_binds,
                ),
            )
            # Start it in the foreground so we wait till it exits (detach=False above)
            volume_task.update(status="Extracting")
            host.client.start(container_pointer)
            host.client.wait(container_pointer["Id"])
            host.client.remove_container(container_pointer["Id"])
            volume_task.update(status="Done", status_flavor=Task.FLAVOR_GOOD)



import attr
import warnings

from ..exceptions import ImageNotFoundException
from ..utils.sorting import dependency_sort


@attr.s
class ContainerFormation:
    """
    Represents a desired or current layout of containers on a single host,
    including their links, image versions, environment etc.

    Formations have a network, which is how we identify which containers might
    be part of a formation or not (e.g. if a container is launched on an
    "eventbrite" network but the formation does not contain it, it should be
    shut down).

    Cross-network linking is done via use of "proxy containers", which get
    put in as a link alias with the right name and forward all connections
    on the appropriate ports to a remote endpoint. Formations are only scoped
    to the host; higher-level management of a set of different hosts is done
    elsewhere.
    """
    graph = attr.ib(repr=False)
    network = attr.ib(default=None)  # a string network name, defaults to graph.prefix
    _instances = attr.ib(default=attr.Factory(list), repr=False)
    container_instances = attr.ib(default=attr.Factory(dict), init=False, repr=False)

    def __attrs_post_init__(self):
        if self.network is None:
            self.network = self.graph.prefix

        for instance in self._instances:
            self.add_instance(instance)

    def validate(self):
        for instance in self:
            instance.validate()

    def add_instance(self, instance):
        """
        Adds an existing instance into the formation.
        """
        assert instance.formation is None
        self.container_instances[instance.name] = instance
        instance.formation = self

    def remove_instance(self, instance):
        """
        Removes an instance from the formation
        """
        # Make sure the instance being removed is part of us
        assert instance.formation is self
        # Resolve the dependent containers so they can all be removed
        dependent_descendency = set(dependency_sort([instance.container], self.graph.dependents)[:-1])
        for other_instance in list(self):
            if other_instance.container in dependent_descendency and other_instance.formation:
                other_instance.formation = None
                del self.container_instances[other_instance.name]
        # Remove the requested container
        del self.container_instances[instance.name]
        instance.formation = None

    def remove_instances(self, instances):
        for instance in instances:
            # Make sure that it was not removed from the formation already as a dependent
            if instance.formation:
                self.remove_instance(instance)

    def add_container(self, container, host):
        """
        Adds a container to run inside the formation along with all dependencies.
        Returns the Instance that was created for the container.
        """
        # Get the list of all dependencies and dependency-ancestors in topological order
        # (this also makes sure there are no cycles as a nice side effect)
        devmodes = self.graph.options(container).get('devmodes', set())
        dependency_ancestry = dependency_sort([container], self.graph.dependencies)[:-1]
        direct_dependencies = self.graph.dependencies(container)
        # Make sure all its dependencies are in the formation
        links = {}
        for dependency in dependency_ancestry:
            # Find the container to satisfy the dependency
            for instance in self:
                if instance.container == dependency:
                    break
            else:
                # OK, we need to make one
                try:
                    instance = self.add_container(dependency, host)
                except ImageNotFoundException as e:
                    # Annotate the error with the container
                    e.container = dependency
                    raise
            if dependency in direct_dependencies:
                links[dependency.name] = instance
        # Look up the image hash to use in the repo
        image_id = host.images.image_version(container.image_name, container.image_tag)
        # Make the instance
        instance = ContainerInstance(
            name="{}.{}.1".format(self.graph.prefix, container.name),
            container=container,
            image_id=image_id,
            links=links,
            devmodes=devmodes,
            foreground=container.foreground,
            environment=container.environment,
        )
        self.add_instance(instance)
        return instance

    def clone(self):
        """
        Clones the formation into a new copy entirely unlinked from this one,
        including new Instances.
        """
        new = self.__class__(self.graph, self.network)
        for instance in self:
            new.add_instance(instance.clone())
        return new

    def has_container(self, container):
        """
        Returns True if the formation has an instance running the given container.
        """
        return any(instance.container == container for instance in self)

    def get_container_instance(self, container_name):
        """
        Given the container name (not the runtime name, the working
        human name to refer to), returns the corresponding
        ContainerInstance.
        """
        for instance in self:
            if instance.container.name == container_name:
                return instance

        raise ValueError("Could not find a running instance of {}".format(container_name))

    def get_instances_using_volume(self, name):
        """
        Return a list of instances that require the named volume.
        """
        return [instance for instance in self if name in instance.container.named_volumes.values()]

    def __getitem__(self, key):
        return self.container_instances[key]

    def __contains__(self, key):
        return key.name in self.container_instances

    def __iter__(self):
        return iter(self.container_instances.values())


@attr.s(hash=False)
class ContainerInstance:
    """
    Represents a single container as part of an overall ContainerFormation
    request.

    :name: The runtime name of the container, like "eventbrite.core-frontend.1"
    :container: The Container instance that's backing this container
    :image_id: The image hash to use for the container
    :links: A dictionary of {alias_str: ContainerInstance} that maps other containers to links
    :devmodes: A set of enabled devmode strings as defined on the Container
    :ports: Exposed ports as {external_port: container_port}
    :environment: Extra environment variables to set in the container
    :command: A custom command override (as a list of string arguments, like subprocess.call takes)
    :foreground: If True, the container is launched in the foreground and a TTY attached
    """

    name = attr.ib(cmp=True)
    container = attr.ib(cmp=False)
    image_id = attr.ib(cmp=False)
    links = attr.ib(default=attr.Factory(dict), repr=False, cmp=False)
    devmodes = attr.ib(default=attr.Factory(set), repr=False, cmp=False)
    ports = attr.ib(default=attr.Factory(dict), repr=False, cmp=False)
    environment = attr.ib(default=attr.Factory(dict), repr=False, cmp=False)
    command = attr.ib(default=None, repr=False, cmp=False)
    foreground = attr.ib(default=None, repr=False, cmp=False)
    formation = attr.ib(default=None, init=False, repr=False, cmp=False)

    def __attrs_post_init__(self):
        self.ports.update(dict(self.container.ports.items()))

    def validate(self):
        """
        Cross-checks the settings we have against the options the Container has
        """
        # Verify all link targets are possible
        for alias, target in list(self.links.items()):
            if isinstance(target, str):
                raise ValueError("Link target {} is still a string!".format(target))
            if target.container not in self.container.graph.dependencies(self.container):
                warnings.warn("It is not possible to link %s to %s as %s" % (target, self.container, alias))
                del self.links[alias]
        # Verify devmodes exist
        for devmode in list(self.devmodes):
            if devmode not in self.container.devmodes:
                warnings.warn("Invalid devmode %s on container %s" % (devmode, self.container.name))
                self.devmodes.remove(devmode)

    def clone(self):
        """
        Returns a safely mutable clone of this Instance
        """
        return self.__class__(
            name=self.name,
            container=self.container,
            image_id=self.image_id,
            links=self.links,
            devmodes=self.devmodes,
            ports=self.ports,
            environment=self.environment,
            command=self.command,
            foreground=self.foreground,
        )

    def different_from(self, other):
        """
        Returns if the other instance is different from this one at all
        (i.e. we need to stop it and start us)
        """
        return (
            self.name != other.name or
            self.container != other.container or
            self.image_id != other.image_id or
            self.links != other.links or
            self.devmodes != other.devmodes or
            self.ports != other.ports or
            self.environment != other.environment or
            self.command != other.command or
            other.foreground or
            self.foreground
        )

    def resolve_links(self):
        """
        Resolves any links that are still names to instances from the formation
        """
        for alias, target in list(self.links.items()):
            # If it's a string, it's come from an introspection process where we couldn't
            # resolve into an instance at the time (as not all of them were around)
            if isinstance(target, str):
                try:
                    target = self.formation[target]
                except KeyError:
                    # We don't error here as that would prevent you stopping orphaned containers;
                    # instead, we delete the link and warn the user. The deleted link means `up` will recreate it
                    # if it's orphaned.
                    del self.links[alias]
                    warnings.warn("Could not resolve link {} to an instance for {}".format(target, self.name))
                else:
                    self.links[alias] = target
            elif isinstance(target, ContainerInstance):
                pass
            else:
                raise ValueError("Invalid link value {}".format(repr(target)))

    def __eq__(self, other):
        return self.name == other.name

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return "<ContainerInstance {} ({})>".format(self.name, self.container.name)



import os
import yaml

from .exceptions import BadConfigError


class Config(object):
    """
    Main config manager. Dumb key-value store that just loads config files
    from various places plus a default and presents a combined view of them.

    This does not handle container configuration or info, just global settings.
    A schema is used to inform how to merge/override values.
    """

    # {section_name: {key: type}}
    schema = {
        "hosts": {
            "*": dict,
        },
        "bay": {
            "home": str,
            "build_log_path": str,
            "user_data_path": str,
            "user_profile_home": str,
            "ssh_agent_container": str,
            "port_proxy_container": str,
        }
    }

    defaults = {
        "hosts": {
            "unix://localhost": {},
        },
        "bay": {
            "home": os.path.expanduser(os.environ.get("BAY_HOME", ".")),
            "build_log_path": os.path.expanduser('~/.bay/{prefix}/build.log'),
            "user_data_path": os.path.expanduser('~/.bay/{prefix}'),
            "user_profile_home": os.path.expanduser('~/.bay'),
            "ssh_agent_container": "tugboat/ssh-agent",
            "port_proxy_container": "tugboat/port-proxy",
        },
    }

    def __init__(self, file_paths):
        self.file_paths = file_paths
        self.load()

    def load(self):
        """
        Loads data from the files.
        """
        self.data = {}
        self.add_config(self.defaults, "<defaults>")
        for file_path in self.file_paths:
            # Load YAML into memory
            with open(file_path, "r") as fh:
                file_data = yaml.safe_load(fh.read())
                self.add_config(file_data, file_path)

    def add_config(self, data, filename):
        """
        Adds the given config on top of the existing ones. Used during load
        and directly for CLI options.
        """
        # Check top level type
        if not isinstance(data, dict):
            raise BadConfigError("Config %s is not a dict at the top level." % filename)
        # Iterate through sections, check type
        for section, items in data.items():
            if not isinstance(items, dict):
                raise BadConfigError("Section %s in %s is not a dict" % (section, filename))
            if section not in self.schema:
                raise BadConfigError("Section %s in %s not in schema" % (section, filename))
            # Iterate through keys, check type
            for key, value in items.items():
                if key not in self.schema[section] and "*" not in self.schema[section]:
                    raise BadConfigError("%s.%s in %s not in schema" % (key, section, filename))
                valid_type = self.schema[section].get(key, self.schema[section].get("*", None))
                assert valid_type is not None
                if not isinstance(value, valid_type):
                    raise BadConfigError("%s.%s in %s is not %s" % (key, section, valid_type))
                # Save value
                self.data.setdefault(section, {})[key] = value

    def __getitem__(self, key):
        return self.data[key]

    def get_path(self, section, key, app):
        """
        Returns the given configuration value, but substituting {prefix} with the
        passed prefix, and making any parent directories. Must be used for file paths.
        """
        value = self.data[section][key].replace("{prefix}", app.containers.prefix)
        dirname = os.path.dirname(value)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)
        return value



import attr
import click

from .base import BasePlugin
from ..cli.argument_types import HostType, ContainerType
from ..exceptions import RegistryRequiresLogin


@attr.s
class RegistryPlugin(BasePlugin):
    """
    Plugin for fetching and uploading images
    """

    def load(self):
        self.add_command(registry)
        self.add_command(push)


@click.group()
def registry():
    """
    Allows operations on registries.
    """
    pass


@registry.command()
@click.option("--host", "-h", type=HostType(), default="default")
@click.pass_obj
def status(app, host):
    """
    Gives registry status
    """
    registry_instance = host.images.get_registry(app)
    if registry_instance is None:
        click.echo("No registry is configured on this project.")
        return
    try:
        url = registry_instance.url(host)
    except RegistryRequiresLogin:
        click.echo("Registry requires login. Run `bay registry login` to do so.")
    else:
        click.echo("Registry configured, docker URL: %s" % url)


@registry.command()
@click.option("--host", "-h", type=HostType(), default="default")
@click.pass_obj
def login(app, host):
    """
    Logs into a registry
    """
    registry_instance = host.images.get_registry(app)
    registry_instance.login(host, app.root_task)


@click.command()
@click.option("--host", "-h", type=HostType(), default="default")
@click.argument("container", type=ContainerType())
@click.argument("tag")
@click.pass_obj
def push(app, host, container, tag):
    """
    Pushes an image up to a registry
    """
    host.images.push_image_version(app, container.image_name, tag, app.root_task)



import attr
import click

from .base import BasePlugin
from ..cli.table import Table
from ..cli.colors import yesno


@attr.s
class HostsPlugin(BasePlugin):
    """
    Plugin for showing information about the current Docker hosts the system
    has.
    """

    def load(self):
        self.add_command(hosts)


@click.command()
@click.pass_obj
def hosts(app):
    """
    Lists available hosts
    """
    # Start a table
    table = Table([
        ("ALIAS", 30),
        ("URL", 50),
        ("VISIBLE", 10),
    ])
    table.print_header()
    # Iterate through docker hosts and ping to see if they're accessible
    for host in app.hosts:
        visible = host.client.ping()
        table.print_row([
            host.alias,
            host.url,
            yesno(visible),
        ])



import attr
import click
from docker.errors import NotFound, APIError
from io import BytesIO
import tarfile

from .base import BasePlugin
from ..cli.argument_types import HostType
from ..cli.table import Table
from ..cli.tasks import Task
from ..docker.introspect import FormationIntrospector


@attr.s
class VolumePlugin(BasePlugin):
    """
    Plugin for showing information about volumes and deleting them.
    """

    provides = ["volume"]
    requires = ["gc"]

    def load(self):
        self.add_command(volume)


@click.group()
def volume():
    """
    Allows operations on volumes.
    """
    pass


@volume.command()
@click.option("--host", "-h", type=HostType(), default="default")
@click.pass_obj
def list(app, host):
    """
    Lists all available volumes
    """
    # Print containers
    table = Table([
        ("NAME", 40),
        ("CONTAINERS", 50)
    ])
    table.print_header()
    # Collect volume information from containers
    users = {}
    for container in app.containers:
        for _, source in container.named_volumes.items():
            users.setdefault(source, set()).add(container.name)
    # Print volumes
    for details in sorted((host.client.volumes()['Volumes'] or []), key=lambda x: x['Name']):
        table.print_row([
            details['Name'],
            ", ".join(users.get(details['Name'], [])),
        ])


@volume.command()
@click.option("--host", "-h", type=HostType(), default="default")
@click.argument("name")
@click.pass_obj
def destroy(app, host, name):
    """
    Destroys a single volume
    """
    task = Task("Destroying volume {}".format(name))
    # Run GC first to clean up stopped containers
    from .gc import GarbageCollector
    GarbageCollector(host).gc_all(task)
    # Remove the volume
    formation = FormationIntrospector(host, app.containers).introspect()
    instance_conflicts = [instance.container.name for instance in formation.get_instances_using_volume(name)]
    if instance_conflicts:
        task.finish(status="Volume {} is in use by container(s): {}".format(
            name, ",".join(instance_conflicts)), status_flavor=Task.FLAVOR_BAD)
    else:
        try:
            host.client.remove_volume(name)
        except NotFound:
            task.add_extra_info("There is no volume called {}".format(name))
            task.finish(status="Not found", status_flavor=Task.FLAVOR_BAD)
        else:
            task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)


@volume.command()
@click.option("--host", "-h", type=HostType(), default="default")
@click.argument("src")
@click.argument("container_name")
@click.argument("volume_name")
@click.pass_obj
def copy_to_docker(app, host, src, container_name, volume_name):
    """
    Copy a local file into docker volumes.
    """
    task = Task("Copying {} to {}:{}".format(src, container_name, volume_name))
    formation = FormationIntrospector(host, app.containers).introspect()

    instance = formation.get_container_instance(container_name)

    # Get the mount path of the volume
    path = instance.container.get_named_volume_path(volume_name)

    # Create a tar stream of the file
    tar_stream = BytesIO()
    with tarfile.open(fileobj=tar_stream, mode="w") as tar:
        tar.add(src)
    tar_stream.seek(0)

    try:
        host.client.put_archive(container=instance.name,
                                path=path,
                                data=tar_stream)
    except APIError:
        task.finish(status="Failed to copy", status_flavor=Task.FLAVOR_BAD)

    else:
        task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)



import attr
import click
import json

from docker.errors import NotFound

from ..cli.tasks import Task
from ..exceptions import ImageNotFoundException, ImagePullFailure, BadConfigError


@attr.s
class ImageRepository:
    """
    A repository of available images for containers.

    Recommended use is internal-only by the plugins, do not use directly.
    """
    host = attr.ib()
    images = attr.ib(default=attr.Factory(dict))

    def list_images(self):
        """
        List all available images.
        """
        raise NotImplementedError()

    def add_image(self, image_name, version, image_hash):
        """
        Add a hash for a given image_name and version.

        This will update any existing hash for an image that was previously
        added to the image repository instance.
        """
        raise NotImplementedError()

    def image_versions(self, image_name):
        """
        Returns a dictionary of version name mapped to the image hash for a
        given image name. May return empty dictionary if there are no images.
        """
        # TODO: Expand to read all tags locally, not just a fixed list
        try:
            return {"latest": self.image_version(image_name, "latest")}
        except ImageNotFoundException:
            return {}

    def get_registry(self, app):
        """
        Given an app, returns the registry handler responsible for handling it
        (or None if it does not need a handler)
        """
        registry = app.containers.registry
        if registry is None:
            return None
        # Work out what plugin to use
        plugin_name, registry_data = registry.split(":", 1)
        # Call the plugin to log in/etc to the registry
        registry_plugins = app.get_catalog_items("registry")
        if plugin_name == "plain":
            # The "plain" plugin is a shortcut for "no plugin"
            return BasicRegistryHandler(app, registry_data)
        elif plugin_name in registry_plugins:
            return registry_plugins[plugin_name](app, registry_data)
        else:
            raise BadConfigError("No registry plugin for {} loaded".format(plugin_name))

    def pull_image_version(self, app, image_name, image_tag, parent_task, fail_silently=False):
        """
        Pulls the most recent version of the given image tag from remote
        docker registry.
        """

        assert isinstance(image_name, str)
        assert isinstance(image_tag, str)

        # The string "local" has a special meaning which means the most recent
        # local image of that name, so we skip the remote call/check.
        if image_tag == "local":
            if fail_silently:
                return None
            else:
                raise ImagePullFailure(
                    "Cannot pull a local image",
                    remote_name=None,
                    image_tag=image_tag
                )

        # See if the registry is willing to give us a URL (it's logged in)
        registry = self.get_registry(app)
        if registry:
            registry_url = registry.url(self.host)
        else:
            registry_url = None
        if registry_url is None:
            if fail_silently:
                return None
            else:
                raise ImagePullFailure(
                    "No registry configured",
                    remote_name=None,
                    image_tag=image_tag
                )

        task = Task(
            "Pulling remote image {}:{}".format(image_name, image_tag),
            parent=parent_task,
            progress_formatter=lambda x: "{} MB".format(x // (1024 ** 2)),
        )

        remote_name = "{registry_url}/{image_name}".format(
            registry_url=registry_url,
            image_name=image_name,
        )

        stream = self.host.client.pull(remote_name, tag=image_tag, stream=True)
        layer_status = {}
        current = None
        total = None
        for line in stream:
            if isinstance(line, bytes):
                line = line.decode("ascii")
            data = json.loads(line)
            if 'error' in data:
                task.finish(status="Failed", status_flavor=Task.FLAVOR_WARNING)
                if fail_silently:
                    return
                else:
                    raise ImagePullFailure(
                        data['error'],
                        remote_name=remote_name,
                        image_tag=image_tag
                    )
            elif 'id' in data:
                if data['status'].lower() == "downloading":
                    layer_status[data['id']] = data['progressDetail']

                elif "complete" in data['status'].lower() and data['id'] in layer_status:
                    layer_status[data['id']]['current'] = layer_status[data['id']]['total']

                if layer_status:
                    statuses = [x for x in layer_status.values()
                                if "current" in x and "total" in x]
                    current = sum(x['current'] for x in statuses)
                    total = sum(x['total'] for x in statuses)

                if total is not None:
                    task.update(progress=(current, total))

        task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)

        # Tag the remote image as the right name
        try:
            self.host.client.tag(
                remote_name + ":" + image_tag,
                image_name,
                tag=image_tag,
                force=True
            )
        except NotFound:
            if fail_silently:
                return
            else:
                raise ImagePullFailure(
                    'Failed to tag {}:{}'.format(remote_name, image_name),
                    remote_name=remote_name,
                    image_tag=image_tag
                )

    def image_version(self, image_name, image_tag):
        """
        Returns the Docker image hash of the requested image and tag, or
        raises ImageNotFoundException if it's not available on the host.
        """
        if image_tag == "local":
            image_tag = "latest"
        try:
            docker_info = self.host.client.inspect_image("{}:{}".format(image_name, image_tag))
            return docker_info['Id']
        except NotFound:
            # TODO: Maybe auto-build if we can?
            raise ImageNotFoundException(
                "Cannot find image {}:{}".format(image_name, image_tag),
                image=image_name,
                image_tag=image_tag,
            )

    def push_image_version(self, app, image_name, image_tag, parent_task):
        """
        Pushes the given image version up to the repository
        """

        assert isinstance(image_name, str)
        assert isinstance(image_tag, str)

        # The string "local" has a special meaning which means the most recent
        # local image of that name, so we skip the remote call/check.
        if image_tag == "local":
            raise ValueError("You cannot push the 'local' version")

        # See if the registry is willing to give us a URL (it's logged in)
        registry = self.get_registry(app)
        if registry:
            registry_url = registry.url(self.host)
        else:
            registry_url = None
        if registry_url is None:
            raise RuntimeError("No registry configured")

        task = Task(
            "Pushing image {}:{}".format(image_name, image_tag),
            parent=parent_task,
            progress_formatter=lambda x: "{} MB".format(x // (1024 ** 2)),
        )

        # Work out the name it needs to be and tag the image as that
        remote_name = "{registry_url}/{image_name}".format(
            registry_url=registry_url,
            image_name=image_name,
        )
        self.host.client.tag(
            image_name + ":" + "latest",
            remote_name,
            tag=image_tag,
            force=True
        )

        # Push it up
        stream = self.host.client.push(remote_name, tag=image_tag, stream=True)
        layer_status = {}
        current = None
        total = None
        for line in stream:
            if isinstance(line, bytes):
                line = line.decode("ascii")
            data = json.loads(line)
            if 'error' in data:
                task.finish(status="Failed", status_flavor=Task.FLAVOR_WARNING)
                raise RuntimeError("Push error: %r" % data['error'])
            elif 'id' in data:
                if data['status'].lower() == "pushing":
                    layer_status[data['id']] = data['progressDetail']
                elif "complete" in data['status'].lower() and data['id'] in layer_status:
                    layer_status[data['id']]['current'] = layer_status[data['id']]['total']
                if layer_status:
                    statuses = [x for x in layer_status.values()
                                if "current" in x and "total" in x]
                    current = sum(x['current'] for x in statuses)
                    total = sum(x['total'] for x in statuses)
                if total is not None:
                    task.update(progress=(current, total))
        task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)


class BasicRegistryHandler:
    """
    Handler for basic (normal Docker) image registries
    """

    def __init__(self, app, data):
        self.registry_url = data

    def url(self, host):
        return self.registry_url

    def login(self, host, task):
        click.echo("Registry does not need a login")

    def logout(self, host, task):
        click.echo("Registry does not need a login")
